/*****************************************************************************
 *
 * FILE:	griffin.c
 * SYNOPSIS:	GPS (Global Positioning System) library
 * DATE:	Thu, Dec 20 2001
 * UPDATED:	Sun, Sep 21 2003
 * AUTHOR:	Kouichi ABE (WALL)
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001-2003 Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (C) 2001-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: griffin.c,v 1.3 2003/09/21 11:16:50 kouichi Exp $
 *
 *****************************************************************************/

#ifdef	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_TERMIOS_H
#include <termios.h>
#endif	/* HAVE_TERMIOS_H */
#if	HAVE_SYSLOG_H
#include <syslog.h>
#endif	/* HAVE_SYSLOG_H */
#include <errno.h>
#include "griffin.h"
#include "gps-sub.h"


/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct {
  int			fd;
  FILE *		fp;
  struct termios	otty;
  GPSDataFormat		type;
} GPS;


/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static GPS	gps;


/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * Function Name:	gpsInit
 * Description:		open the GPS device and set the file pointer
 * Arguments:		data format, device name
 * Return Value:	successful: 0, otherwise: -1
 */
int
gpsInit(data_format, gps_dev)
	GPSDataFormat	data_format;
	const char *	gps_dev;
{
  struct termios	tty;
  speed_t		speed;

  /* initialize */
  gps.fd = -1;
  gps.fp = NULL;

  /* set GPS device */
  if (gps_dev == NULL) {
    gps_dev = GPS_DEVICE;
  }

  /* set terminal attributes */
  gps.fd = open(gps_dev, O_RDONLY);
  if (gps.fd < 0) {
    syslog(LOG_ERR, "%s: %m", gps_dev);
    return -1;
  }
  if (tcgetattr(gps.fd, &tty) < 0) {
    syslog(LOG_ERR, "%s: %m", gps_dev);
    close(gps.fd);
    return -1;
  }
  gps.otty = tty;

  /* set communication speed */
  switch (data_format) {
    case GDF_SONY_PACY:
      speed = PACY_OPENSPEED;
      break;
    case GDF_NMEA0183:
      speed = NMEA_OPENSPEED;
      break;
    default:
      speed = B4800;
      break;
  }
  gps.type = data_format;
  if (cfsetspeed(&tty, speed) < 0) {
    syslog(LOG_ERR, "%s: %m", gps_dev);
    close(gps.fd);
    return -1;
  }

  /* set attributes */
  tty.c_cflag |= (tty.c_cflag & ~CSIZE) | CS8;	/* 8bit char */
  tty.c_cflag &= ~(PARENB|PARODD);		/* non-parity */
#if	0
  tty.c_iflag |= IXON | IXOFF;		/* software flow on */
#else
  tty.c_iflag &= ~(IXON|IXOFF|IXANY);	/* software flow off */
#endif
  tty.c_cflag &= ~CRTSCTS;		/* hardware flow off */

#if	1	/* UNUSED?	*/
  /* set into raw, no echo mode */
#if	1	/* for Windows */
  tty.c_iflag |= ICRNL;
#endif
#if	LINUX
  tty.c_iflag &= ~ICRNL;	/* for Linux */
#endif	/* LINUX */
  tty.c_iflag |= IGNBRK;
  tty.c_lflag = 0;
  tty.c_oflag = 0;
  tty.c_cflag |= (CLOCAL | CREAD);
#endif	/* UNUSED?	*/
  if (tcsetattr(gps.fd, TCSANOW, &tty) < 0) {
    syslog(LOG_ERR, "%s: %m", gps_dev);
    close(gps.fd);
    return -1;
  }

  /* set the file pointer for the GPS device */
  gps.fp = fdopen(gps.fd, "r");
  if (gps.fp  == NULL) {
    syslog(LOG_ERR, "%s: %m", gps_dev);
    close(gps.fd);
    return -1;
  }

  if (data_format == GDF_SONY_PACY) {
    char	dummy[GBUFSIZE];
    GPSData	data;

    fgets(dummy, GBUFSIZE, gps.fp);	/* take away first incomplete line. */
    if (pacy_getinfo(&data.gps.pacy, gps.fp) &&
	strcmp(data.gps.pacy.version, "SONY81") != 0) {
      syslog(LOG_ERR, "Griffin supports only SONY GPS ver. 8.1");
      fclose(gps.fp);
      return -1;
    }
  }

  return 0;
}

/*
 * Function Name:	gpsQuit
 * Description:		close the file pointer for GPS device
 * Arguments:		none
 * Return Value:	none
 */
void
gpsQuit(void)
{
  if (gps.fp != NULL) {
    int	ret;

    if (tcsetattr(gps.fd, TCSANOW, &gps.otty) == -1) {
      ;	/* ignore */
    }

    while (((ret = fclose(gps.fp)) < 0) && (errno == EINTR)) {
      ;
    }
    gps.fp = NULL;
    gps.fd = -1;
  }
}

/*
 * Function Name:	gpsGetInfo
 * Description:		get information from GPS receiver
 * Arguments:		gps_info structure
 * Return Value:	Successful: 0, Otherwise: -1
 */
int
gpsGetInfo(gpsi)
	register GPSInfo *	gpsi;
{
  GPSData	data;
  int		n = -1;

  memset((char *)&data, 0, sizeof(data));
  switch (gps.type) {
    case GDF_SONY_PACY:
      n = pacy_getinfo(&data.gps.pacy, gps.fp);
      if (n == 0) {
	pacy_raw2std(&data.gps.pacy, gpsi);
      }
      break;
    case GDF_NMEA0183:
      if (nmea_getinfo(&data.gps.nmea, gps.fp) == 0) {
	n = nmea_raw2std(&data.gps.nmea, gpsi);
      }
      break;
    default:
      break;
  }

  return n;
}

/*
 * Function Name:	gpsPrintInfo
 * Description:		show the GPS information
 * Arguments:		gps_info structure, output sink
 * Return Value:	none
 */
void
gpsPrintInfo(gpsi, fout)
	register GPSInfo *	gpsi;
	register FILE *		fout;
{
  fputs(asctime(&gpsi->timestamp), fout);
  fprintf(fout, "(%c%02d'%02d\"%02d.%d, %c%03d'%02d\"%02d.%d)\n",
		gpsi->ns,
		gpsi->latitude.degrees,
		gpsi->latitude.minutes,
		gpsi->latitude.seconds,
		gpsi->latitude.fraction,
		gpsi->ew,
		gpsi->longitude.degrees,
		gpsi->longitude.minutes,
		gpsi->longitude.seconds,
		gpsi->longitude.fraction);
  fprintf(fout, "%.1f[km/h] %.1f[deg] %.1f[m]\n",
		gpsi->velocity,
		gpsi->direction,
		gpsi->altitude);
  fprintf(fout, "dop[%.1f] status[%c] satellite[%d]\n",
		gpsi->dop,
		gpsi->status,
		gpsi->satellite);
}

